-- bindings.lua
-- handle override bindings for ActionSlots
-- the main challenge with action slot bindings is that the actual command
-- associated with a binding may change in the event a slot becomes a movement
-- slot, etc so we can't rely upon simply looking up the command associated
-- with a button

local Addon = _G[...]
local ToFunctionArgs = Addon.Util.ToFunctionArgs
local tinsert = table.insert
local tconcat = table.concat
local strsplit = strsplit
local strjoin = strjoin

local BindingsManager = {}

function BindingsManager:OnLoad()
    self.keyFrames = {}

    self.owner = CreateFrame("Frame", nil, nil, "SecureHandlerBaseTemplate")
    self.owner:Hide()

    self.owner:Execute([[
        Bindings = table.new()
        Frames = table.new()
    ]])

    self.owner:SetAttribute("ClearBindings", [[
        wipe(Bindings)
        self:ClearBindings()
    ]])

    self.owner:SetAttribute("ClearFrameBindings", [[
        -- ClearFrameBindings
        local frameName = ...
        local touched = false

        for key, name in pairs(Bindings) do
            if name == frameName then
                touched = true

                Bindings[key] = nil
                self:SetBinding(false, key)
                self:CallMethod("OnSetBinding", key)
            end
        end

        if touched then
            self:CallMethod("OnFrameBindingsUpdated", frameName)
        end
    ]])

    self.owner:SetAttribute("SetBinding", [[
        -- SetBinding
        local key, frameName = ...
        local oldFrameName = Bindings[key]
        local action = self:RunAttribute("GetAction", frameName)

        if frameName ~= oldFrameName then
            Bindings[key] = frameName
            self:SetBinding(false, key, action)
            self:CallMethod("OnSetBinding", key, frameName)

            if oldFrameName ~= nil then
                self:CallMethod("OnFrameBindingsUpdated", oldFrameName)
            end

            if frameName ~= nil then
                self:CallMethod("OnFrameBindingsUpdated", frameName)
            end
        end
    ]])

    self.owner:SetAttribute("UpdateFrameBindings", [[
        -- UpdateFrameBindings
        local frameName = ...
        local action = self:RunAttribute("GetAction", frameName)

        for key, name in pairs(Bindings) do
            if name == frameName then
                self:SetBinding(false, key, action)
            end
        end
    ]])

    self.owner:SetAttribute("GetAction", [[
        -- GetAction
        local frameName = ...
        local frame = Frames[frameName]

        if frame then
            local action = frame:GetAttribute("bindaction")
            if action and action ~= "" then
                return action
            end

            local virtualButton = frame:GetAttribute("virtualButton")
            if virtualButton and virtualButton ~= "" then
                return ("CLICK %s:LeftButton"):format(virtualButton)
            end
        end

        return ("CLICK %s:LeftButton"):format(frameName)
    ]])

    self.owner:SetAttribute("AddFrame", [[
        local frame = self:GetFrameRef("addFrame")
        local frameName = frame:GetName()

        if not Frames[frameName] then
            Frames[frameName] = frame
            self:RunAttribute("UpdateFrameBindings", frameName)
        end
    ]])

    self.owner:SetAttribute("DelFrame", [[
        local frame = self:GetFrameRef("delFrame")
        local frameName = frame:GetName()

        if Frames[frameName] then
            Frames[frameName] = nil

            for key, name in pairs(Bindings) do
                if name == frameName then
                    self:SetBinding(false, key, nil)
                end
            end
        end
    ]])

    self.owner.OnSetBinding = function(_, key, frameName)
        self:OnSetBinding(key, frameName)
    end

    self.owner.OnFrameBindingsUpdated = function(_, frameName)
        local frame = _G[frameName]
        if frame then
            Addon:Notify("OnFrameBindingsUpdated", frame)
        end
    end

    Addon.RegisterCallback(self, "OnActionSlotLoaded")
    Addon.RegisterCallback(self, "OnActionSlotActivated")
    Addon.RegisterCallback(self, "OnActionSlotDeactivated")
end

function BindingsManager:OnActionSlotLoaded(msg, slot)
    slot:SetFrameRef("BindingsManager", self.owner)

    local virtualButton = Addon.VirtualButton:Create(slot)
    slot:SetAttribute("virtualButton", virtualButton:GetName())
end

function BindingsManager:OnActionSlotActivated(msg, slot)
    self:AddFrame(slot)
end

function BindingsManager:OnActionSlotDeactivated(msg, slot)
    self:RemoveFrame(slot)
end

function BindingsManager:OnSetBinding(key, frameName)
    local oldFrameName = self.keyFrames[key]

    if oldFrameName == frameName then
        return
    end

    -- tag the binding with this frame
    self.keyFrames[key] = frameName

    -- remove binding from old frame
    if oldFrameName ~= nil then
        local oldBindings = self.settings[oldFrameName]
        local bindings = {}

        for k in oldBindings:gmatch("([^%s]+)") do
            if k ~= key then
                tinsert(bindings, k)
            end
        end

        if #bindings > 0 then
            self.settings[oldFrameName] = tconcat(bindings, " ")
        else
            self.settings[oldFrameName] = ""
        end
    end

    -- add binding to new frame
    if frameName ~= nil then
        local newBindings = self.settings[frameName]

        if newBindings == "" then
            newBindings = key
        else
            newBindings = strjoin(" ", newBindings, key)
        end

        self.settings[frameName] = newBindings
    end
end

function BindingsManager:Initialize(settings)
    self.settings = settings

    self.owner:Execute([[ self:RunAttribute("ClearBindings") ]])
    wipe(self.keyFrames)

    for frameName, bindings in pairs(settings) do
        for key in bindings:gmatch("([^%s]+)") do
            local cmd = [[ local key, name = %s; Bindings[key] = name ]]

            self:SecureExecute(cmd, key, frameName)
            self.keyFrames[key] = frameName
        end
    end
end

function BindingsManager:SetBinding(key, frameName)
    self:SecureExecute([[ self:RunAttribute("SetBinding", %s) ]], key, frameName)
end

function BindingsManager:ClearFrameBindings(frameName)
    self:SecureExecute([[ self:RunAttribute("ClearFrameBindings", %s) ]], frameName)
end

function BindingsManager:GetFrameBindings(frameName)
    local bindings = self.settings[frameName]

    if bindings ~= "" then
        return strsplit(" ", bindings)
    end
end

function BindingsManager:AddFrame(frame)
    self.owner:SetFrameRef("addFrame", frame)
    self.owner:Execute([[ self:RunAttribute("AddFrame") ]])
end

function BindingsManager:RemoveFrame(frame)
    self.owner:SetFrameRef("delFrame", frame)
    self.owner:Execute([[ self:RunAttribute("DelFrame") ]])
end

function BindingsManager:SecureExecute(cmd, ...)
    self.owner:Execute(cmd:format(ToFunctionArgs(...)))
end

BindingsManager:OnLoad()

Addon.BindingsManager = BindingsManager
