local Addon = _G[...]
local MacroSlot = Addon.ActionSlots:CreateKind("macro")
local GetItemIDFromString = Addon.Util.GetItemIDFromString
local IsAttackSpellID = Addon.Util.IsAttackSpellID
local IsAutoRepeatSpellID = Addon.Util.IsAutoRepeatSpellID
local IsSpellIDInRange = Addon.Util.IsSpellIDInRange

local function withSpellID(with, without)
    if type(without) == "function" then
        return function(self)
            local spellID = self:GetSpellID()

            if spellID then
                return with(spellID)
            end

            return without(self)
        end
    end

    return function(self)
        local spellID = self:GetSpellID()

        if spellID then
            return with(spellID)
        end
    end
end

local function withItemID(with, without)
    if type(without) == "function" then
        return function(self)
            local itemID = self:GetItemID()

            if itemID then
                return with(itemID)
            end

            return without(self)
        end
    end

    return function(self)
        local itemID = self:GetItemID()

        if itemID then
            return with(itemID)
        end
    end
end

local function withItemOrSpellID(withItem, withSpell, without)
    return withItemID(withItem, withSpellID(withSpell, without))
end

-- charges, maxCharges, chargeStart, chargeDuration
MacroSlot.GetActionCharges = withSpellID(GetSpellCharges, function() return 0, 0, 0, 0, 0 end)

MacroSlot.GetActionCooldown = withItemOrSpellID(GetItemCooldown, GetSpellCooldown, function() return 0, 0, 0, 0 end)

MacroSlot.GetActionCount = withItemOrSpellID(GetItemCount, GetSpellCount, function() return 0 end)

function MacroSlot:GetActionID()
    return 0
end

function MacroSlot:GetActionInfo()
    return "macro", self:GetAttribute("macro")
end

MacroSlot.GetActionLossOfControlCooldown = withSpellID(GetSpellLossOfControlCooldown, function() return 0, 0 end)

function MacroSlot:GetActionText()
    return _G.GetMacroInfo(self:GetAttribute("macro")) or ""
end

function MacroSlot:GetActionTexture()
    local _, icon = _G.GetMacroInfo(self:GetAttribute("macro"))
    return icon
end

function MacroSlot:GetItemID()
    local macroID = self:GetAttribute("macro")
    if macroID then
        local itemString = select(2, _G.GetMacroItem(macroID))
        if itemString then
            return GetItemIDFromString(itemString)
        end
    end
end

function MacroSlot:GetSpellID()
    local macroID = self:GetAttribute("macro")
    if macroID then
        return _G.GetMacroSpell(macroID)
    end

    local itemID = self:GetItemID()
    if itemID then
        local _, spellID = _G.GetItemSpell(itemID)
        return spellID
    end
end

function MacroSlot:HasAction()
    return true
end

MacroSlot.IsActionInRange = withItemOrSpellID(IsItemInRange, IsSpellIDInRange)

MacroSlot.IsAttackAction = withSpellID(IsAttackSpellID)

MacroSlot.IsAutoCastAction = withSpellID(GetSpellAutocast)

MacroSlot.IsAutoRepeatAction = withSpellID(IsAutoRepeatSpellID)

MacroSlot.IsConsumableAction = withItemOrSpellID(IsConsumableItem, IsConsumableSpell)

MacroSlot.IsCurrentAction = withItemOrSpellID(IsCurrentItem, IsCurrentSpell)

MacroSlot.IsEquippedAction = withItemID(IsEquippedItem)

MacroSlot.IsItemAction = withItemID(function() return true end)

MacroSlot.IsStackableAction = withItemID(function(itemID)
    local stackCount = select(8, _G.GetItemInfo(itemID))

    return stackCount and stackCount > 1
end)

MacroSlot.IsUsableAction = withItemOrSpellID(
    IsUsableItem,
    IsUsableSpell,
    function() return true end
)

function MacroSlot:SetTooltip(tooltip)
    local macroID = self:GetAttribute("macro")

    if macroID then
        local spellID = self:GetSpellID()
        if spellID then
            return tooltip:SetSpellByID(spellID)
        end

        local itemID = self:GetItemID()
        if itemID then
            return tooltip:SetItemByID(itemID)
        end

        local name = _G.GetMacroInfo(macroID)
        if name then
            return GameTooltip_SetTitle(tooltip, name, HIGHLIGHT_FONT_COLOR)
        end
    end
end

Addon.MacroSlot = MacroSlot