local Addon = _G[...]

local BarStates = {}

local states = {}

-- setup default state map
function BarStates:OnLoad()
    local function add(stateType, stateId, stateValue, stateText)
        return self:Add {
            type = stateType,
            id = stateId,
            value = stateValue,
            text = stateText
        }
    end

    local function newFormConditionLookup(spellID)
        return function()
            for i = 1, GetNumShapeshiftForms() do
                local _, _, _, formSpellID = GetShapeshiftFormInfo(i)

                if spellID == formSpellID then
                    return ('[form:%d]'):format(i)
                end
            end
        end
    end

    --keybindings
    add('modifier', 'selfcast', '[mod:SELFCAST]', AUTO_SELF_CAST_KEY_TEXT)
    add(
        'modifier',
        'ctrlAltShift',
        '[mod:alt,mod:ctrl,mod:shift]',
        strjoin('-', CTRL_KEY_TEXT, ALT_KEY_TEXT, SHIFT_KEY_TEXT)
    )
    add('modifier', 'ctrlAlt', '[mod:alt,mod:ctrl]', strjoin('-', CTRL_KEY_TEXT, ALT_KEY_TEXT))
    add('modifier', 'altShift', '[mod:alt,mod:shift]', strjoin('-', ALT_KEY_TEXT, SHIFT_KEY_TEXT))
    add('modifier', 'ctrlShift', '[mod:ctrl,mod:shift]', strjoin('-', CTRL_KEY_TEXT, SHIFT_KEY_TEXT))
    add('modifier', 'alt', '[mod:alt]', ALT_KEY)
    add('modifier', 'ctrl', '[mod:ctrl]', CTRL_KEY)
    add('modifier', 'shift', '[mod:shift]', SHIFT_KEY)

    --paging
    for i = 2, 6 do
        add('page', 'page' .. i, ('[bar:%d]'):format(i), _G['BINDING_NAME_ACTIONPAGE' .. i])
    end

    --class
    do
        local class = select(2, UnitClass('player'))

        if class == 'DRUID' then
            add('class', 'moonkin', '[bonusbar:4]', GetSpellInfo(24858))
            add('class', 'bear', '[bonusbar:3]', GetSpellInfo(5487))
            add('class', 'tree', newFormConditionLookup(33891), GetSpellInfo(33891))
            add('class', 'prowl', '[bonusbar:1,stealth]', GetSpellInfo(5215))
            add('class', 'cat', '[bonusbar:1]', GetSpellInfo(768))
        elseif class == 'ROGUE' then
            add('class', 'shadowdance', '[form:2]', GetSpellInfo(185313))
            add('class', 'stealth', '[bonusbar:1]', GetSpellInfo(1784))
        elseif class == 'PALADIN' then
            add('class', 'crusader', newFormConditionLookup(32223), GetSpellInfo(32223))
            add('class', 'devotion', newFormConditionLookup(465), GetSpellInfo(465))
            add('class', 'retribution', newFormConditionLookup(183435), GetSpellInfo(183435))
            add('class', 'concentration', newFormConditionLookup(317920), GetSpellInfo(317920))
        end

        local race = select(2, UnitRace('player'))
        if race == 'NightElf' then
            add('class', 'shadowmeld', '[stealth]', GetSpellInfo(58984))
        end
    end

    --target reaction
    add('target', 'help', '[help]')
    add('target', 'harm', '[harm]')
    add('target', 'notarget', '[noexists]')

    --group
    add('group', 'raid', '[group:raid]')
    add('group', 'party', '[group:party]')
    add('group', 'group', '[group]')

    --automatic updating for UPDATE_SHAPESHIFT_FORMS
    do
        local f = CreateFrame('Frame')

        f:SetScript(
            'OnEvent',
            function()
                Addon.Callbacks:Fire('OnShapeshiftFormsUpdated')
            end
        )

        f:RegisterEvent('UPDATE_SHAPESHIFT_FORMS')

        f:Hide()
    end
end

function BarStates:Add(state, index)
    if index then
        return tinsert(states, index, state)
    end

    return tinsert(states, state)
end

function BarStates:Get(id)
    for _, v in pairs(states) do
        if v.id == id then
            return v
        end
    end
end

function BarStates:GetAll()
    return ipairs(states)
end

function BarStates:Count()
    return #states
end

do
    local function typesIterator()
        local type = nil

        return function(t, i)
            for j = i + 1, #t do
                local state = t[j]

                if state and state.type ~= type then
                    return j, type
                end
            end
        end
    end

    function BarStates:GetTypes()
        return typesIterator(), states, 0
    end
end

do
    local function stateTypesIterator(type)
        return function(t, i)
            for j = i + 1, #t do
                local state = t[j]

                if state and state.type ~= type then
                    return j, type
                end
            end
        end
    end

    function BarStates:GetAllOfType(type)
        return stateTypesIterator(type), states, 0
    end
end

do
    local function filterIterator(func, ...)
        return function(t, i)
            for j = i + 1, #t do
                local state = t[j]
                if state and func(state) then
                    return j, state
                end
            end
        end
    end

    function BarStates:GetFiltered(func)
        return filterIterator(func), states, 0
    end
end

BarStates:OnLoad()

-- exports
Addon.BarStates = BarStates
