local Addon = _G[...]
local GamePad = CreateFromMixins({}, Addon.Bar)
local ActionSlots = Addon.ActionSlots

-- events
function GamePad:OnConstruct()
    Addon.Bar.OnConstruct(self)

    self.regions = {}
    self.activeRegions = {}
    self.buttons = {}
    self.activeButtons = {}
end

function GamePad:OnInitialize()
    Addon.Bar.OnInitialize(self)

    self.currentSpec = GetSpecialization() or 1

    for id, button in pairs(self.buttons) do
        button:Initialize(self:GetButtonSettings(id))
    end

    self:Layout()
    self:UpdateRegionStateDrivers()
end


function GamePad:OnOverrideBarUpdated()
    if InCombatLockdown() then return end

    for region in pairs(self.activeRegions) do
        region:Execute([[
            self:RunAttribute("UpdateOverrideOffset")
            self:RunAttribute("UpdateState")
        ]])
    end
end

-- commands
function GamePad:Layout()
    local width = 0
    local height = 0
    local prev
    local pW, pH = self:GetPadding()
    local rSpacing = self:GetRegionSpacing()
    local regions = self:GetDeviceLayout()

    for region in pairs(self.activeRegions) do
        self.activeRegions[region] = false
    end

    for button in pairs(self.activeButtons) do
        self.activeButtons[button] = false
    end

    for id, layout in ipairs(regions) do
        local region = self:GetOrCreateRegion(id)

        self:LayoutRegion(region, layout)
        self.activeRegions[region] = true

        if prev then
            region:SetPoint("TOP", prev, "BOTTOM", 0, -rSpacing)
            height = height + region:GetHeight() + rSpacing
        else
            region:SetPoint("TOP", 0, -pH)
            height = height + region:GetHeight()
        end

        width = max(width, region:GetWidth())

        prev = region
    end

    for region, active in pairs(self.activeRegions) do
        if not active then
            region:Hide()
        end
    end

    for button, active in pairs(self.activeButtons) do
        if active then
            button:Activate()
        else
            button:Deactivate()
        end
    end

    self.frame:SetSize(width + pW*2, height + pH*2)
end

function GamePad:LayoutRegion(region, layout)
    if layout.shape == Addon.RegionShapes.Dpad then
        self:LayoutDpadRegion(region, layout)
    else
        self:LayoutGridRegion(region, layout)
    end
end

function GamePad:LayoutDpadRegion(region, layout)
	local spacing = self:GetSpacing()
    local cols = 4
	local bSize = _G.ActionButton1:GetWidth() + spacing
    local r = (bSize * (cols - 1)) / 2
    local offset = (r - bSize/cols) - spacing

    for i = 1, #layout.buttons - 2, 1 do
        local id = layout.buttons[i]

        if id then
            local button = self:GetOrCreateButton(id)
            button:ClearAllPoints()
            button:SetParent(region)

            local a = 360 * ((i - 1) / cols)
            local x = offset * cos(a)
            local y = offset * sin(a)

            button:SetPoint("CENTER", x, y)
            self.activeButtons[button] = true
        end
    end

    local tlButtonID = layout.buttons[#layout.buttons - 1]
    if tlButtonID then
        local button = self:GetOrCreateButton(tlButtonID)
        button:ClearAllPoints()
        button:SetParent(region)
        button:SetPoint("TOPLEFT")
        self.activeButtons[button] = true
    end

    local brButtonID = layout.buttons[#layout.buttons]
    if brButtonID then
        local button = self:GetOrCreateButton(brButtonID)
        button:ClearAllPoints()
        button:SetParent(region)
        button:SetPoint("BOTTOMRIGHT")
        self.activeButtons[button] = true
    end

    region:SetSize(2 * (r + bSize) - spacing, 2 * (r + bSize) - bSize - spacing*2)
end

function GamePad:LayoutGridRegion(region, layout)
    local cols = layout.columns
	local rows = ceil(#layout.buttons / cols)
	local spacing = self:GetSpacing()
    local buttonWidth = _G.ActionButton1:GetWidth() + spacing
    local buttonHeight = _G.ActionButton1:GetHeight() + spacing

    for i, id in pairs(layout.buttons) do
        local button = self:GetOrCreateButton(id)
        local row = floor((i - 1) / cols)
        local col = (i - 1) % cols
        local x = buttonWidth * col
        local y = buttonHeight * row

        button:SetParent(region)
        button:ClearAllPoints()
        button:SetPoint('TOPLEFT', x, -y)
        self.activeButtons[button] = true
	end

	local barWidth = (buttonWidth * cols) - spacing
	local barHeight = (buttonHeight * rows) - spacing
	region:SetSize(barWidth, barHeight)
end

function GamePad:GetOrCreateRegion(id)
    local region = self.regions[id]

    if not region then
        region = CreateFrame("Frame", nil, self.frame, "SecureHandlerStateTemplate")
        region.id = id

        region:SetAttribute("_onstate-force", [[
            self:RunAttribute("UpdateState")
        ]])

        region:SetAttribute("_onstate-page", [[
            self:RunAttribute("UpdateOverrideOffset")
            self:RunAttribute("UpdateState")
        ]])

        region:SetAttribute("_onstate-override", [[
            self:RunAttribute("UpdateOverrideOffset")
            self:RunAttribute("UpdateState")
        ]])

        region:SetAttribute("UpdateState", [[
            local oldState = self:GetAttribute("currentState")
            local newState

            local forceState = self:GetAttribute("state-force")
            if forceState then
                newState = forceState
            else
                local overrideIndex = self:GetAttribute("overrideIndex") or 0
                if overrideIndex > 120 then
                    newState = "override"
                else
                    newState = self:GetAttribute("state-page") or "default"
                end
            end

            if oldState ~= newState then
                self:SetAttribute("currentState", newState)
                control:ChildUpdate("currentState", newState)
            end
        ]])

        region:SetAttribute("UpdateOverrideOffset", [[
            local overrideState = self:GetAttribute("state-override") or 0
            local oldIndex = self:GetAttribute("overrideIndex") or 0
            local newIndex = 1

            if overrideState ~= 0 then
                if HasVehicleActionBar() then
                    newIndex = GetVehicleBarIndex() or 1
                elseif HasOverrideActionBar() then
                    newIndex = GetOverrideBarIndex() or 1
                elseif HasTempShapeshiftActionBar() then
                    newIndex = GetTempShapeshiftBarIndex() or 1
                else
                    newIndex = 1
                end
            else
                newIndex = 1
            end

            if oldIndex ~= newIndex then
                local offset = (newIndex - 1) * 12
                self:SetAttribute("overrideIndex", offset)
                control:ChildUpdate("overrideIndex", offset)
            end
        ]])

        self.regions[id] = region
    end

    return region
end

function GamePad:GetOrCreateButton(id)
    local button = self.buttons[id]

    if not button then
        button = ActionSlots:GetOrCreate(id)
        button:Initialize(self:GetButtonSettings(id))

        self.buttons[id] = button
    end

    return button
end

function GamePad:GetButtonSettings(id)
    return Addon:GetActionSlotSettings(id)
end

function GamePad:GetDeviceLayout()
    return Addon.DeviceLayouts[self:GetDeviceType()]
end

function GamePad:UpdateRegionStateDrivers()
    for _, region in pairs(self.regions) do
        self:UpdateRegionStateDriver(region)
        self:UpdateRegionOverrideStateDriver(region)
    end
end

local function getCondition(state)
    if type(state.value) == 'function' then
        return state.value()
    end

    return state.value
end

do
    local buffer = {}
    local twipe = wipe
    local tinsert = table.insert
    local tconcat = table.concat

    function GamePad:UpdateRegionStateDriver(region)
        UnregisterStateDriver(region, 'page', "default")

        if not self.activeRegions[region] then
            return
        end

        twipe(buffer)

        for _, state in Addon.BarStates:GetAll() do
            local page = self:GetRegionStateMap(region.id, state.id)
            if page > 0 then
                local condition = getCondition(state)
                if condition then
                    tinsert(buffer, ("%spage%d"):format(condition, page))
                end
            end
        end

        tinsert(buffer, "1")

        if #buffer > 1 then
            RegisterStateDriver(region, "page", tconcat(buffer, ";"))
        end
    end

    function GamePad:UpdateRegionOverrideStateDriver(region)
        RegisterStateDriver(region, "override", "[overridebar][possessbar][vehicleui][@vehicle,exists][shapeshift][form]1;0")
    end
end

-- properties
local function getValidOption(table, value)
    for _, v in pairs(table) do
        if v == value then
            return v
        end
    end

    return select(2, next(table))
end

function GamePad:SetDeviceMode(mode)
    local oldMode = self:GetDeviceMode()
    local newMode = getValidOption(Addon.DeviceModes, mode)

    if oldMode ~= newMode then
        self.sets.deviceMode = newMode
        self:Layout()

        Addon:Notify("OnBarPropertyChanged", self, "deviceMode", newMode)
    end
end

function GamePad:GetDeviceMode()
    return self.sets.deviceMode or Addon.DeviceModes.Automatic
end

function GamePad:SetDeviceType(id)
    local oldType = self:GetDeviceType()
    local newType = getValidOption(Addon.DeviceTypes, id)

    if oldType ~= newType then
        self.sets.deviceType = newType
        self:Layout()

        Addon:Notify("OnBarPropertyChanged", self, "deviceType", newType)
    end
end

function GamePad:GetDeviceType()
    if self:GetDeviceMode() == Addon.DeviceModes.Automatic then
        if RazerGamingKeypad_ActiveDevice and RazerGamingKeypad_ActiveDevice == 2 then
            return Addon.DeviceTypes.Tartarus
        end

        return Addon.DeviceTypes.Orbweaver
    end

    return self.sets.deviceType or Addon.DeviceTypes.Orbweaver
end

function GamePad:SetPadding(width, height)
    local oldWidth, oldHeight = self:GetPadding()
    local newWidth = tonumber(width) or 0
    local newHeight = tonumber(height) or newWidth

    if not (oldWidth == newWidth and oldHeight == newHeight) then
        self.sets.padW = newWidth
        self.sets.padH = newHeight

        Addon.Callbacks:Fire("OnBarPropertyChanged", self, "padding", newWidth, newHeight)

        self:Layout()
    end
end

function GamePad:GetPadding()
    local w = tonumber(self.sets.padW) or 0
    local h = tonumber(self.sets.padH) or w

    return w, h
end

function GamePad:SetRegionSpacing(spacing)
    local oldSpacing = self:GetRegionSpacing()
    local newSpacing = tonumber(spacing) or 0

    if oldSpacing ~= newSpacing then
        self.sets.regionSpacing = newSpacing
        self:Layout()

        Addon.Callbacks:Fire("OnBarPropertyChanged", self, "regionSpacing", newSpacing)
    end
end

function GamePad:GetRegionSpacing()
    return tonumber(self.sets.regionSpacing) or 0
end

function GamePad:SetSpacing(spacing)
    local oldSpacing = self:GetSpacing()
    local newSpacing = tonumber(spacing) or 0

    if oldSpacing ~= newSpacing then
        self.sets.spacing = newSpacing

        Addon.Callbacks:Fire("OnBarPropertyChanged", self, "spacing", newSpacing)

        self:Layout()
    end
end

function GamePad:GetSpacing()
    return tonumber(self.sets.spacing) or 0
end

function GamePad:SetRegionStateMap(regionId, stateId, pageId)
    local oldPage = self:GetRegionStateMap(regionId, stateId)
    local newPage = tonumber(pageId) or -1

    if oldPage ~= newPage then
        self.sets.regionStateMaps[regionId][stateId] = newPage
        self:UpdateRegionStateDriver(self.regions[regionId])

        Addon.Callbacks:Fire("OnBarPropertyChanged", self, "regionStateMaps", regionId, stateId, newPage)
    end
end

function GamePad:GetRegionStateMap(regionId, stateId)
    return self.sets.regionStateMaps[regionId][stateId] or -1
end

function GamePad:ForcePage(pageID)
    if InCombatLockdown() then return end

    for region in pairs(self.activeRegions) do
        region:SetAttribute("state-force", pageID)
    end
end

Addon.GamePad = GamePad