local AddonName = ...
local Addon = _G[AddonName]
local L = LibStub("AceLocale-3.0"):GetLocale(AddonName)

local GamePadMenu = {}

function GamePadMenu:OnLoad()
    self.menu = nil
    self.menuShown = false

    Addon.RegisterCallback(self, "OnToggleBarOptionsMenu")
end

function GamePadMenu:OnToggleBarOptionsMenu(msg, bar, show)
    if bar.id ~= "gamePad" then return end

    if show then
        self:ShowMenu(bar)
    else
        self:HideMenu(bar)
    end
end

function GamePadMenu:ShowMenu(bar)
    local menu = self.menu or self:CreateMenu(bar)

    menu:Hide()
    menu:SetOwner(bar)
    menu:ShowPanel("General")
    menu:Show()
end

function GamePadMenu:HideMenu(bar)
    local menu = self.menu
    if menu then
        menu:Hide()
    end
end

function GamePadMenu:CreateMenu(bar)
    local menu = Addon.Config.BarMenu:New()

    self:AddGeneralPanel(menu)
    self:AddRegionPanel(menu, 1, L.MainPad)
    self:AddRegionPanel(menu, 2, L.DPad)

    self.menu = menu
    return menu
end

function GamePadMenu:AddGeneralPanel(menu)
    local panel = menu:NewPanel(L.General)

    panel:NewDropdown{
		name = L.DeviceType,

        get = function()
            if panel.owner:GetDeviceMode() == Addon.DeviceModes.Automatic then
                return "auto"
            else
                return panel.owner:GetDeviceType()
            end
        end,

        set = function(_, value)
            if value == "auto" then
                panel.owner:SetDeviceMode(Addon.DeviceModes.Automatic)
            else
                panel.owner:SetDeviceMode(Addon.DeviceModes.Manual)
                panel.owner:SetDeviceType(value)
            end
        end,

		items = {
            {text = L.Automatic, value = "auto"},
            {text = L.RazerOrbweaver, value = 1},
            {text = L.RazerTartarus, value = 2}
		}
    }

    panel:NewHeader(L.Layout)

    panel.spacingSlider = panel:NewSpacingSlider()

    panel.regionSpacingSlider = panel:NewSlider{
		name = L.RegionSpacing,
		min = -16,
		max = 32,

		get = function()
			return panel.owner:GetRegionSpacing()
		end,

		set = function(_, value)
			panel.owner:SetRegionSpacing(value)
		end
    }

    panel.paddingSlider = panel:NewPaddingSlider()

	panel.scaleSlider = panel:NewScaleSlider()

    panel.opacitySlider = panel:NewOpacitySlider()

    panel.fadeSlider = panel:NewFadeSlider()

    return panel
end

function GamePadMenu:AddRegionPanel(menu, regionId, name)
    local panel = menu:NewPanel(name)

    local function getDropdownItems()
        local prevNumPages = nil
        local items = {}

        return function()
		    local numPages = Addon:NUM_GAMEPAD_PAGES()

            if prevNumPages ~= numPages then
                prevNumPages = numPages

                wipe(items)

			    items = {
                    { value = -1, text = _G.DISABLE },
			    }

			    for i = 1, numPages do
				    items[i + 1] = {
					    value = i,
					    text = L.PageX:format(i)
				    }
                end
            end

            return items
		end
    end

    local function getLabel(state)
        local n = state.text

        if type(n) == 'function' then
            n = n()
        elseif not n then
            n = L["BarState_" .. state.id]
        end

        return n
    end

    local type = nil
    for _, state in Addon.BarStates:GetAll() do
        if state.type ~= type then
            panel:NewHeader(L["BarStateType_" .. state.type])
            type = state.type
        end

        panel:NewDropdown{
            name = getLabel(state),

            items = getDropdownItems(),

            get = function()
                return panel.owner:GetRegionStateMap(regionId, state.id)
            end,

            set = function(_, value)
                panel.owner:SetRegionStateMap(regionId, state.id, value)
            end
        }
    end

    return panel
end

GamePadMenu:OnLoad()

Addon.GamePadMenu = GamePadMenu