local AddonName, AddonData = ...
local Addon = LibStub("AceAddon-3.0"):NewAddon(AddonData, AddonName, "AceEvent-3.0", "AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(AddonName)
local KeyBound = LibStub("LibKeyBound-1.0")

local CONFIG_ADDON_NAME = AddonName .. "Config"
local CHAR_KEY = strjoin(" - ", UnitName("player"), GetRealmName())


Addon.Callbacks = LibStub("CallbackHandler-1.0"):New(Addon)
Addon:SetDefaultModuleLibraries("AceEvent-3.0", "AceConsole-3.0")

-- events
function Addon:OnEnable()
	self:InitializeDatabase()
	self:SetupConfigAddonLoader()
	self:SetupLauncher()

    self.currentSpec = GetSpecialization() or 1

    self.BindingsManager:Initialize(self.db.profile.bindings)

    self.gamePad = self.GamePad:GetOrCreate("gamePad", "Game Pad")
    self.gamePad:Initialize(self.db.profile)

	self:RegisterEvent("ADDON_LOADED", "OnAddonLoaded")
    self:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
	self:RegisterEvent('UPDATE_BONUS_ACTIONBAR', 'OnOverrideBarUpdated')
	self:RegisterEvent('UPDATE_VEHICLE_ACTIONBAR', 'OnOverrideBarUpdated')
	self:RegisterEvent('UPDATE_OVERRIDE_ACTIONBAR', 'OnOverrideBarUpdated')

	if _G.RazerNaga then
		_G.RazerNaga.Envoy:Register(self, "CONFIG_MODE_UPDATE", "OnNagaConfigModeUpdated")
	end
end

function Addon:OnAddonLoaded(msg, addon)
	if addon == CONFIG_ADDON_NAME then
		self:UnregisterEvent(msg)

		self.configAddonLoaded = true
		self.Callbacks:Fire("OnConfigAddonLoaded", self.Config)
	end
end

function Addon:OnNagaConfigModeUpdated(msg, enabled)
	self:SetConfigurationMode(enabled)
end

function Addon:OnOverrideBarUpdated()
    self.gamePad:OnOverrideBarUpdated()
end

function Addon:OnProfileLoaded()
	if not self.gamePad then return end

    self.BindingsManager:Initialize(self.db.profile.bindings)
    self.gamePad:Initialize(self.db.profile)
end

function Addon:PLAYER_SPECIALIZATION_CHANGED(msg, unit, ...)
    if unit ~= "player" then return end

    local oldSpec = self.currentSpec
    local newSpec = GetSpecialization()

    if oldSpec ~= newSpec then
        self.currentSpec = newSpec
        self.gamePad:Initialize(self.db.profile)
    end
end

function Addon:GetActionSlotSettings(id)
	return self.db.profile.actionSlotContents[CHAR_KEY][self.currentSpec][id]
end

function Addon:SetUseAsOverrideBar(enable)
    self.db.profile.useAsOverrideBar = enable and true or false
end

function Addon:UsingAsOverrideBar()
    return self.db.profile.useAsOverrideBar
end

function Addon:ForcePage(pageID)
    self.gamePad:ForcePage(pageID)
end

function Addon:NUM_GAMEPAD_PAGES()
    return self.db.global.numGamepadPages
end


-- messaging
function Addon:Notify(msg, ...)
	self.Callbacks:Fire(msg, ...)
end

-- load the options menu when the interface options frame is loaded
function Addon:SetupConfigAddonLoader()
	local loader = CreateFrame("Frame", nil, _G.InterfaceOptionsFrame)

	loader:SetScript(
		"OnShow",
		function()
			self:LoadConfigAddon()
			loader:SetScript("OnShow", nil)
		end
	)
end

function Addon:SetupLauncher()
	local nagaLauncher = LibStub("LibDataBroker-1.1"):GetDataObjectByName("RazerNaga")

	if nagaLauncher then
		local gamepadTitle = select(2, GetAddOnInfo(AddonName))
		local nagaTitle = select(2, GetAddOnInfo("RazerNaga"))
		local launcherTitle = strjoin(" + ", nagaTitle, gamepadTitle)

		hooksecurefunc(nagaLauncher, "OnTooltipShow", function(tooltip)
			local title = _G[tooltip:GetName() .. 'TextLeft1']
			if title then
				title:SetText(launcherTitle)
			end
		end)
	else
		local iconPath = ([[Interface\Addons\%s\images\logo_icon]]):format(AddonName)

		local dataObject =
			LibStub("LibDataBroker-1.1"):NewDataObject(
			AddonName,
			{
				type = "launcher",
				icon = iconPath,
				OnClick = function(_, button)
					if button == "LeftButton" then
						if IsShiftKeyDown() then
							self:ToggleBindingMode()
						else
							self:ToggleConfigurationMode()
						end
					elseif button == "RightButton" then
						self:ShowOptionsMenu()
					end
				end,
				OnTooltipShow = function(tooltip)
					if not tooltip or not tooltip.AddLine then
						return
					end

					local title = select(2, GetAddOnInfo(AddonName))
					GameTooltip_SetTitle(tooltip, title)
					GameTooltip_AddInstructionLine(tooltip, L.LauncherLeftClickInstructions)
					GameTooltip_AddInstructionLine(tooltip, L.LauncherShiftLeftClickInstructions)
					GameTooltip_AddInstructionLine(tooltip, L.LauncherRightClickInstructions)
				end
			}
		)

		LibStub("LibDBIcon-1.0"):Register(AddonName, dataObject, self.db.global.launcher)
	end
end

function Addon:LoadConfigAddon()
	if not self.configAddonLoaded then
		return _G.LoadAddOn(CONFIG_ADDON_NAME)
	end

	return true
end

-- options menu display
function Addon:ShowOptionsMenu()
	if self:LoadConfigAddon() then
		self.Callbacks:Fire("OnShowOptionsMenu")
	end
end

-- minimap button display
function Addon:SetShowLauncher(enable)
	self.db.global.launcher.hide = not enable
	LibStub('LibDBIcon-1.0'):Refresh(AddonName, self.db.global.launcher)
end

function Addon:ShowingLauncher()
	return not self.db.global.launcher.hide
end

-- binding mode
function Addon:IsBindingModeEnabled()
	return KeyBound:IsShown()
end

function Addon:ToggleBindingMode()
	KeyBound:Toggle()
end

-- configuration mode
do
	local enabled = false

	function Addon:InConfigurationMode()
		return enabled
	end

	function Addon:SetConfigurationMode(enable)
		enable = enable and true or false

		if enabled ~= enable then
			enabled = enable

			if enable then
				self:LoadConfigAddon()
			end

			local lock = not enable
			if _G.RazerNaga and _G.RazerNaga:Locked() ~= lock then
				_G.RazerNaga:SetLock(lock)
			end

			self.Callbacks:Fire("OnConfigurationModeChanged", enable)
		end
	end

	function Addon:ToggleConfigurationMode()
		self:SetConfigurationMode(not self:InConfigurationMode())
	end
end

-- bar options menu display
do
	local shownBar = nil

	function Addon:ToggleBarOptionsMenu(bar)
		local oldBar = shownBar
		local newBar = bar
		shownBar = nil

		if oldBar then
			self.Callbacks:Fire("OnToggleBarOptionsMenu", oldBar, false)
		end

		if newBar and newBar ~= oldBar and self:LoadConfigAddon() then
			self.Callbacks:Fire("OnToggleBarOptionsMenu", newBar, true)
		end
	end
end

-- toggles
do
	local toggles = {
		showEmptyActionButtons = {
			get = "ShowingEmptyActionButtons",
			set = "SetShowEmptyActionButtons"
		},
		showActionButtonBindings = {
			get = "ShowingActionButtonBindings",
			set = "SetShowActionButtonBindings"
		},
		showActionButtonLabels = {
			get = "ShowingActionButtonLabels",
			set = "SetShowActionButtonLabels"
		},
		showActionButtonEquippedItems = {
			get = "ShowingActionButtonEquippedItemBorders",
			set = "SetShowActionButtonEquippedItemBorders"
		},
		showActionButtonTooltips = {
			get = "ShowingActionButtonTooltips",
			set = "SetShowActionButtonTooltips"
		},
		showActionButtonTooltipsInCombat = {
			get = "ShowingActionButtonTooltipsInCombat",
			set = "SetShowActionButtonTooltipsInCombat"
		}
	}

	local function toboolean(value)
		if value then
			return true
		end
		return false
	end

	for key, setting in pairs(toggles) do
		local kind = setting.kind or "profile"

		Addon[setting.set] = function(self, value)
			local enabled = toboolean(value)

			if enabled ~= self[setting.get](self) then
				self.db[kind][key] = enabled
				self.Callbacks:Fire("OnSettingChanged", key, enabled)
			end
		end

		Addon[setting.get] = function(self)
			return self.db[kind][key]
		end
	end
end

-- exports
_G[AddonName] = Addon
