local AddonName = ...
local Addon = _G[AddonName]

local select = select
local tconcat = table.concat
local tonumber = tonumber
local tostring = tostring
local twipe = wipe
local type = type

local Util = {}

function Util.Memoize(func)
    return setmetatable({}, {
        __index = function(t, k)
            local v = func(k)

            if v ~= nil then
                t[k] = v
            end

            return v
        end
    })
end

-- retrieves the last index of ... that isn't null
function Util.GetLastNonNilIndex(...)
    for i = select("#", ...), 1, -1 do
        if select(i, ...) ~= nil then
            return i
        end
    end
    return 0
end

-- converts a variable argument list into a string that looks like a list
-- of function parameters. this is for usage with secure state calls
do
    local temp = {}

    local function toFunctionArg(v)
        if type(v) == "nil" then
            return "nil"
        end

        if type(v) == "string" then
            return ("%q"):format(v)
        end

        return tostring(v)
    end

    local getLastNonNilIndex = Util.GetLastNonNilIndex

    function Util.ToFunctionArgs(...)
        local count = getLastNonNilIndex(...)

        if count > 0 then
            twipe(temp)

            for i = 1, count do
                temp[i] = toFunctionArg(select(i, ...))
            end

            return tconcat(temp, ", ")
        end
    end
end

-- cache lookups from itemStrings to IDs
do
    local itemIDs = setmetatable({}, {__index = function(t, k)
        local itemString = k
        local id = tonumber(itemString)

        if not id then
            id = itemString and tonumber(itemString:match("item:(%d+)"))
        end

        if id then
            t[k] = id
            return id
        end
    end})

    function Util.GetItemIDFromString(itemString)
        if itemString ~= nil then
            return itemIDs[itemString]
        end
    end
end

-- most, but not all Spell functions support spellIDs
-- wrap the ones that don't so that we have a consistent API
do
    local GetSpellInfo = GetSpellInfo
    local IsAttackSpell = IsAttackSpell
    local IsAutoRepeatSpell = IsAutoRepeatSpell
    local IsSpellInRange = IsSpellInRange

    local function toSpellIDTest(nameTest)
        return function(spellID)
            local name = (GetSpellInfo(spellID))

            if name then
                local result = nameTest(name)
                local resultType = type(result)

                if resultType == "number" then
                    return result == 1
                end

                if resultType ~= "nil" then
                    return result
                end
            end
        end
    end
    -- if performance is an issue, we probably want to memoize this one
    Util.IsAttackSpellID = toSpellIDTest(IsAttackSpell)

    -- and this one
    Util.IsAutoRepeatSpellID = toSpellIDTest(IsAutoRepeatSpell)

    -- but not this one
    Util.IsSpellIDInRange = toSpellIDTest(IsSpellInRange)
end

-- A functional way to fade a frame from one opacity to another without
-- constantly creating new animation groups for the frame
do
    local faders = {}

    local function fader_OnFinished(self)
        self:GetParent():SetAlpha(self.toAlpha)

        if type(self.onFinished) == "function" then
            self.onFinished(self:GetParent())
        end
    end

    local function fader_Create(parent)
        local animation = parent:CreateAnimationGroup()
        animation:SetLooping('NONE')
        animation:SetScript('OnFinished', fader_OnFinished)

        local fade = animation:CreateAnimation('Alpha')
        fade:SetSmoothing('IN_OUT')

        return function(toAlpha, duration, onFinished)
            if animation:IsPlaying() then
                animation:Pause()
            end

            fade:SetFromAlpha(parent:GetAlpha())
            fade:SetToAlpha(toAlpha)
            fade:SetDuration(duration)

            animation.toAlpha = toAlpha
            animation.onFinished = onFinished
            animation:Play()
        end
    end

    function Util.Fade(frame, toAlpha, duration, onFinished)
        local fade = faders[frame]

        if not fade then
            fade = fader_Create(frame)
        end

        fade(toAlpha, duration, onFinished)
    end
end

Addon.Util = Util